"""cloud_sptheme.ext.relbar_toc - addes TOC entry to relbar"""
import os.path
import logging; log = logging.getLogger(__name__)

def insert_toc(app, pagename, templatename, ctx, event_arg):
    links = ctx['rellinks']

    #remove any existing toc (present on some pages)
    for idx,  elem in enumerate(links):
        if elem[3] == "toc":
            del links[idx]
            break

    #place toc right after "next" / "previous"
    idx = -1
    for idx, entry in enumerate(links):
        if entry[3] in ("next","previous"):
            break
    else:
        idx += 1

    #insert our toc entry
    #FIXME: there's probably a MUCH better / less broken way to do this
    path = os.path.split(os.path.splitext(ctx['pathto']("contents"))[0])[1]
    ##path = os.path.splitext(ctx['pathto']("contents"))[0]
    ##if path == '':
    ##    path = pagename
    links.insert(idx, (path, "Table Of Contents", "C", "toc"))

def setup(app):
    app.connect('html-page-context', insert_toc)



"""cloud_sptheme.utils -- internal helper utilities"""
#=============================================================================
# imports
#=============================================================================
# core
from functools import update_wrapper
import logging; log = logging.getLogger(__name__)
import sys
# site
# pkg
# local
__all__ = [
    # py2/3 compat
    'PY2', 'PY3', 'u', 'ru',

    # monkeypatch helpers
    "patchapplier",
    "monkeypatch",
]

#=============================================================================
# internal py2/3 compat helpers
#=============================================================================
PY2 = sys.version_info < (3,0)
PY3 = not PY2

# FIXME: these aren't very rigorous / correct, but they work for current purposes.
if PY2:
    def u(s):
        return s.decode("unicode_escape")
    def ru(s):
        return s.decode("ascii")
else:
    def u(s):
        return s
    ru = u

#=============================================================================
# monkeypatch helpers
#=============================================================================
def patchapplier(func):
    """
    function decorator to help functions that apply a monkeypatch.
    makes them only run once.
    """
    def wrapper():
        if wrapper.patched:
            return False
        func()
        wrapper.patched = True
        logging.getLogger(func.__module__).debug("%s: patch applied", func.__name__)
        return True
    wrapper.patched = False
    update_wrapper(wrapper, func)
    return wrapper

def monkeypatch(target, name=None):
    """
    helper to monkeypatch another object.
    the decorated function is wrapped around the existing function in
    :samp:`target.{name}`, and used to replace it.

    **name** defaults to the name of the function being decorated.

    the original value is passed in as the first positional argument to the function.
    """
    def builder(func):
        attr = name or func.__name__
        wrapped = getattr(target, attr)
        def wrapper(*args, **kwds):
            return func(wrapped, *args, **kwds)
        update_wrapper(wrapper, wrapped)
        wrapper.__wrapped__ = wrapped # not set by older update_wrapper() versions
        setattr(target, attr, wrapper)
        return func # return orig func so we can use it again
    return builder


#=============================================================================
# eof
#=============================================================================



"""cloud_sptheme.ext.table_styling -- add directives for styling tables"""
#=============================================================================
# imports
#=============================================================================
# core
from cloud_sptheme.utils import PY3
if PY3:
    from itertools import zip_longest as izip_longest
else:
    # FIXME: not present in py25
    from itertools import izip_longest
import os
from shutil import copyfile
# site
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.tables import RSTTable
from sphinx.builders.html import StandaloneHTMLBuilder
# pkg
from cloud_sptheme import _root_dir, is_cloud_theme
# local
__all__ = [
    "setup",
]

#=============================================================================
# constants
#=============================================================================

# name of key controlling whether css file is included
EMBED_KEY = "table_styling_embed_css"

# name of key controlling css class name
CLASS_KEY = "table_styling_class"

#=============================================================================
# field option parsers
#=============================================================================
def _split_argument_list(argument):
    if "," in argument:
        return argument.split(",")
    else:
        return argument.split()

def _parse_argument_map(argument, argmap, param):
    args = _split_argument_list(argument)
    if len(args) == 1 and all(c in argmap for c in args[0]):
        args = args[0]
    def norm(arg):
        try:
            return argmap[arg]
        except KeyError:
            raise ValueError("invalid %s: %r" % (param, arg))
    return [norm(arg) for arg in args]

_alignment_map = dict(
    l="left",
    r="right",
    c="center",
    j="justify",
    left="left",
    right="right",
    center="center",
    centered="center", # compat alias
    justify="justify",
    justified="justify", # compat alias
)

def alignment_list(argument):
    """convert into list of alignment options.
    raise ``ValueError`` if no args found, or invalid strings.
    """
    return _parse_argument_map(argument, _alignment_map, "alignment")

_bool_map = {"true": True, "t": True,
             "yes": True, "y": True,
             "false": False, "f": False,
             "no": False, "n": False,
             }

def bool_list(argument):
    """convert to list of true/false values"""
    return _parse_argument_map(argument, _bool_map, "boolean value")

def class_option_list(argument):
    """convert to list of list of classes"""
    args = _split_argument_list(argument)
    return [directives.class_option(arg) for arg in args]

_divider_map = {
    "0": "no",
    "1": "single",
    "2": "double",
    "none": "no",
    "single": "single",
    "double": "double",
    }

def divider_list(argument):
    return _parse_argument_map(argument, _divider_map, "divider style")

#=============================================================================
# replacement for table directive
#=============================================================================
class ExtendedRSTTable(RSTTable):
    # TODO: could have this auto-generate tabularcolumns directive for latex,
    #       based on alignment, dividers, and wrapping settings:
    #           |,|| for dividers
    #           l,c,r - align, nowrap
    #           L,C,R,J - align, wrap
    #       XXX: how to handle widths for tabularcolumns?

    option_spec = RSTTable.option_spec.copy()
    option_spec.update({
        # class, name already present
        ##'header-rows': directives.nonnegative_int,
        'header-columns': directives.nonnegative_int,
        # TODO: column-widths: support limited set of units (em/in/%)
        #       expressable under both css & latex
        'widths': directives.positive_int_list,
        'column-alignment': alignment_list,
        'column-wrapping': bool_list,
        'column-classes': class_option_list,
        'column-dividers': divider_list,
    })

    def run(self):
        result = RSTTable.run(self)
        if result and isinstance(result[0], nodes.table):
            self._update_table_classes(result[0])
        return result

    def _update_table_classes(self, table):
        assert isinstance(table, nodes.table)
        config = self.state.document.settings.env.config
        classes = getattr(config, CLASS_KEY)
        if classes:
            for cls in classes.split():
                table.set_class(cls)
        header_cols = self.options.get("header-columns") or 0
##        header_rows = self.options.get("header-rows")
        widths = self.options.get("widths")
        dividers = self.options.get("column-dividers")
        if dividers is None:
            get_divider = None
        else:
            def get_divider(idx):
                try:
                    return dividers[idx]
                except IndexError:
                    return "no"
        EMPTY = ()
        opts = (
            self.options.get("column-alignment", EMPTY),
            self.options.get("column-wrapping", EMPTY),
            self.options.get("column-classes", EMPTY),
        )
        def locate(cls):
            for child in table.children:
                if isinstance(child, cls):
                    return child
            return None
        tgroup = locate(nodes.tgroup)
        if not tgroup:
            return
        col = 0
        for child in tgroup:
            if isinstance(child, nodes.colspec):
                if widths and col < len(widths):
                    child['colwidth'] = widths[col]
                if col < header_cols:
                    child['stub'] = 1
                col += 1
                continue
            assert isinstance(child, (nodes.thead, nodes.tbody))
            for row in child.children:
                # add alignment and wrap classes to each entry (would add to
                # colspec, but html doesn't inherit much from colgroup)
                assert isinstance(row, nodes.row)
                for idx, (entry, align, wrap, clist) in \
                  enumerate(izip_longest(row, *opts)):
                    if entry is None:
                        # FIXME: make into propert rst error
                        raise ValueError("not enough columns for field options")
                    assert isinstance(entry, nodes.entry)
                    classes = entry['classes']
                    if align:
                        classes.append(align + "-align")
                    if wrap is False:
                        classes.append("nowrap")
                    if clist:
                        classes.extend(clist)
                    if get_divider:
                        classes.append(get_divider(idx) + "-left-divider")
                        classes.append(get_divider(idx+1) + "-right-divider")
        # untested - might be missing some docutils node framework bits
        ##if header_rows > 1:
        ##    thead = locate(nodes.thead)
        ##    if not thead:
        ##        thead = nodes.thead()
        ##        idx = table.children.index(tgroup)
        ##        table.children.insert(idx, thead)
        ##    thead.extend(tgroup.children[:header_rows])
        ##    del tgroup.children[:header_rows]

#=============================================================================
# patch builder to copy css file (if needed)
#=============================================================================
def prepare_builder(app):
    # make sure needed css styling gets included when building html
    builder = app.builder
    if not isinstance(builder, StandaloneHTMLBuilder):
        return
    value = getattr(app.config, EMBED_KEY)
    if value is None:
        value = not is_cloud_theme(app.config.html_theme)
    if not value:
        return

    # add custom css stylesheet
    name = "table_styling.css"
    app.add_stylesheet(name)

    # monkeypatch builder to copy over css file
    orig = builder.copy_static_files
    def wrapper():
        orig()
        source = os.path.join(_root_dir, "ext", name)
        target = os.path.join(builder.outdir, "_static", name)
        copyfile(source, target)
    builder.copy_static_files = wrapper

#=============================================================================
# register extension
#=============================================================================
def setup(app):
    # whether it will embed raw css
    app.add_config_value(EMBED_KEY, None, "html")

    # default class to add to all styled tables
    app.add_config_value(CLASS_KEY, "styled-table", "html")

    # replace existing table directive with custom one
    app.add_directive("table", ExtendedRSTTable)

    # add extra resources
    app.connect("builder-inited", prepare_builder)

#=============================================================================
# eof
#=============================================================================



"""
cloud_sptheme.ext.docfield_markup -- monkeypatches sphinx to allow ``~`` in docfields.
"""
#=============================================================================
# imports
#=============================================================================
# core
import logging; log = logging.getLogger(__name__)
# site
# pkg
from cloud_sptheme.utils import patchapplier, monkeypatch
# local
__all__ = [
    "setup",
]

#=============================================================================
# patch
#=============================================================================
@patchapplier
def _patch_docfield():
    from sphinx.util.docfields import Field, nodes, addnodes

    # NOTE: would like to just wrap make_xref(), but have to override so
    #       many parts that just end up replicating all the code :(
    #       hence this ignored _wrapped() entirely

    @monkeypatch(Field)
    def make_xref(_wrapped, self, rolename, domain, target, innernode=nodes.emphasis, contnode=None):
        # 'contnode' argument added in sphinx 1.3 
        # not sure what it does, aborting if feature is present
        if contnode:
            return _wrapped(self, rolename, domain, target, innernode=innernode, contnode=contnode)
        # NOTE: this tries to replicate the convention used in PyXRefRole.process_link()
        rawtext = title = target
        if rawtext.startswith("~"):
            # if the first character is a tilde, don't display the module/class
            # parts of the contents
            target = target.lstrip("~")
            title = target.rpartition(".")[2]
        if issubclass(innernode, nodes.Text):
            # Text classes want rawtext second
            node = innernode(title, rawtext)
        else:
            # Element classes want rawtext first
            node = innernode(rawtext, title)
        if not rolename:
            return node
        refnode = addnodes.pending_xref(title, refdomain=domain, refexplicit=True,
                                        reftype=rolename, reftarget=target)
        refnode += node
        return refnode

#=============================================================================
# sphinx entrypoint
#=============================================================================
def setup(app):
    # don't apply our patch unless actually loaded by sphinx
    _patch_docfield()

#=============================================================================
# eoc
#=============================================================================



"""
cloud_sptheme setup script
"""
#=========================================================
#init script env
#=========================================================
import sys,os
from os.path import abspath, join
root_path = abspath(join(__file__, ".."))
os.chdir(root_path)
lib_path = '.'
#=========================================================
#imports
#=========================================================
import re
from setuptools import setup, find_packages
import subprocess
import time
#=========================================================
#inspection
#=========================================================
with open(os.path.join("cloud_sptheme", "__init__.py")) as fh:
    version = re.search(r"""(?m)^__version__\s*=\s*['"](.*?)['"]\s*$""", fh.read()).group(1)
if version.endswith((".dev0", ".post0")):
    stamp = time.strftime("%Y%m%d%H%M%S")
    if os.path.exists(".hg"):
        try:
            stamp = subprocess.check_output(["hg", "tip", "-T", "{date(date, '%Y%m%d%H%M%S')}+r{node|short}"]).decode("utf-8")
        except (OSError, subprocess.CalledProcessError):
            pass
    version = version[:-1] + stamp

#=========================================================
#setup
#=========================================================
setup(
    #package info
    packages = find_packages(),
    package_data = { "cloud_sptheme": ["themes/*/*.*", "themes/*/static/*.*", "ext/*.css"] },
    zip_safe=False,

    install_requires=[ "sphinx>=1.2"],

    # metadata
    name = "cloud_sptheme",
    version = version,
    author = "Eli Collins",
    author_email = "elic@assurancetechnologies.com",
    description = "a nice sphinx theme named 'Cloud', and some related extensions",
    long_description="""\
This is a small package containing a Sphinx theme named "Cloud",
along with some related Sphinx extensions. To see an example
of the theme in action, check out it's documentation
at `<http://packages.python.org/cloud_sptheme>`_.
    """,
    license = "BSD",
    keywords = "sphinx extension theme",
    url = "https://bitbucket.org/ecollins/cloud_sptheme",
    download_url = "http://pypi.python.org/pypi/cloud_sptheme",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        # XXX: there should be a "Framework :: Sphinx :: Extension" classifier :)
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Documentation',
        'Topic :: Software Development :: Documentation',
    ],
    entry_points={
        'sphinx_themes': [
            'path = cloud_sptheme:get_theme_dir',
        ],
    },
)
#=========================================================
#EOF
#=========================================================
